#include <stdio.h>
#include <stdbool.h>
#include <stddef.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <pthread.h>
#include <signal.h>
#include <sched.h>
#include <sys/ioctl.h>
#include <sys/stat.h>

#include <time.h>

#include <linux/rpmsg.h>

#define NUM_ITERATIONS (10)
#define RPMSG_MASTER_ADDR (0x1001)
#define RPMSG_REMOTE_ADDR (0x4004)

static int g_efd = 0, g_local_endpt = 0;

/* Returns a pointer to allocated memory, needs to be freed once done */
char *file_deref_link(char *fpath, char *link_name)
{
	char path[512] = { 0 };
	char rel_path[256] = { 0 };
	int n, nr;

	n = snprintf(path, 256, "%s/%s", fpath, link_name);
	if (n < 0 || n >= 256) {
		fprintf(stderr, "%s: could not create full path string\n",
			__func__);
		return NULL;
	}

	nr = readlink(path, rel_path, sizeof(rel_path));
	if (nr < 0) {
		fprintf(stderr, "%s: readlink failed for %s\n", __func__, path);
		return NULL;
	}

	if (n + nr >= 512) {
		fprintf(stderr, "%s: full relative path exceeds buffer size, n = %d nr = %d\n",
			__func__, n, nr);
		return NULL;
	}

	memset(path, 0, sizeof(path));
	sprintf(path, "%s/%s", fpath, rel_path);

	return realpath(path, NULL);
}

static int file_read_string(char *fpath, char *buf, int size)
{
	int fd, bytes;

	fd = open(fpath, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "could not open %s: errno = %d\n",
			fpath, errno);
		return -errno;
	}

	bytes = read(fd, buf, size);
	close(fd);
	if (bytes <= 0) {
		fprintf(stderr, "could not read %s: errno = %d\n",
			fpath, errno);
		return -EIO;
	}
	if (bytes >= size) {
		fprintf(stderr, "%d bytes read from %s are larger than size %d\n",
			bytes, fpath, size);
		return -EIO;
	}

	/* suppress the newline */
	buf[bytes - 1] = '\0';

	return bytes;
}

static int file_read_value(char *fpath)
{
	char buf[32];
	int ret;

	ret = file_read_string(fpath, buf, sizeof(buf));
	if (ret < 0)
		return ret;

	return strtol(buf, NULL, 0);
}

static int rpmsg_char_get_local_endpt(void)
{
	char fpath[512] = { 0 };
	char rpmsg[16] = { 0 };
	char *rpath;

	sprintf(rpmsg, "rpmsg%u", 0);
	rpath = file_deref_link("/sys/class/rpmsg", rpmsg);
	if (!rpath) {
		fprintf(stderr, "%s: rpmsg0 realpath failed\n",
			__func__);
		return -ENOENT;
	}

	sprintf(fpath, "%s/src", rpath);
	free(rpath);

	return file_read_value(fpath);
}

static int rpmsg_char_destroy_eptdev(int fd)
{
	int ret;

	ret = ioctl(fd, RPMSG_DESTROY_EPT_IOCTL, NULL);
	if (ret) {
		fprintf(stderr, "%s: could not destroy endpt %d\n",
			__func__, fd);
		return ret;
	}

	close(fd);
	return 0;
}

static int rpmsg_char_open_eptdev(char *eptdev_name, int flags)
{
	int efd, endpt;

	efd = open("/dev/rpmsg0", O_RDWR | flags);
	if (efd < 0) {
		fprintf(stderr, "failed to open eptdev /dev/rpmsg0\n");
		return -errno;
	}

	/* local end-point is assigned after the open call */
	endpt = rpmsg_char_get_local_endpt();
	if (endpt < 0) {
		rpmsg_char_destroy_eptdev(efd);
		fprintf(stdout, "%s: get_local_endpt failed %d \n",__func__, endpt);
		return endpt;
	}

	g_efd = efd;
	g_local_endpt = endpt;

	return 0;
}

static int rpmsg_char_create_eptdev(char *eptdev_name,
				     int local_port, int remote_port)
{
	int fd, ret;
	struct rpmsg_endpoint_info ept_info = { 0 };

	fd = open("/dev/rpmsg_ctrl0", O_RDWR);
	if (fd < 0) {
		fprintf(stderr, "%s: could not open rpmsg_ctrl0 dev node\n", __func__);
		return fd;
	}

	ept_info.src = local_port;
	ept_info.dst = remote_port;
	sprintf(ept_info.name, "%s", eptdev_name);

	ret = ioctl(fd, RPMSG_CREATE_EPT_IOCTL, &ept_info);
	if (ret) {
		fprintf(stderr, "%s: ctrl_fd ioctl: %s\n", __func__,
			strerror(errno));
	}

	close(fd);

	return ret;
}

int send_msg(int fd, char *msg, int len)
{
	int ret = 0;

	ret = write(fd, msg, len);
	if (ret < 0) {
		perror("Can't write to rpmsg endpt device\n");
		return -1;
	}

	return ret;
}

int recv_msg(int fd, int len, char *reply_msg, int *reply_len)
{
	int ret = 0;

	/* Note: len should be max length of response expected */
	ret = read(fd, reply_msg, len);
	if (ret < 0) {
		perror("Can't read from rpmsg endpt device\n");
		return -1;
	} else {
		*reply_len = ret;
	}

	return 0;
}

void usage(void)
{
	printf("Usage: rpmsg_rw [-n <num_msgs>] \n");
	printf("\t\tDefaults: num_msgs: %d\n", NUM_ITERATIONS);
}

struct timespec startTime, endTime;
#define NSEC_PER_SEC (1000000000L)

#define DIFF_NS(A, B) (((B).tv_sec - (A).tv_sec) * NSEC_PER_SEC + (B).tv_nsec - (A).tv_nsec)

int main(int argc, char const *argv[])
{
	char packet_buf[512] = { 0 };
	int num_msgs = NUM_ITERATIONS;
	int ret, i, packet_len, c;
	uint64_t time_ns, total_time = 0, avg_time, max_time = 0;

	while (1) {
		c = getopt(argc, (void *)argv, "n:h");
		if (c == -1)
			break;

		switch (c) {
		case 'n':
			num_msgs = atoi(optarg);
			break;
		case 'h':
		default:
			usage();
			exit(0);
		}
	}

	ret = rpmsg_char_create_eptdev("rpmsg-char-0", RPMSG_MASTER_ADDR, RPMSG_REMOTE_ADDR);
	if (ret) {
		fprintf(stderr, "%s: could not create end-point rpmsg-char-0\n", __func__);
		return ret;
	}

	ret = rpmsg_char_open_eptdev("rpmsg-char-0", 0);
	if (ret) {
		fprintf(stderr, "%s: could not open end-point rpmsg-char-0\n", __func__);
		return ret;
	}

	for (i = 0; ; i++)  {
	//while (1) {
		memset(packet_buf, 0, sizeof(packet_buf));
		sprintf(packet_buf, "hello there %d!", i);
		packet_len = strlen(packet_buf);
		printf("Sending message #%d: %s\n", i, packet_buf);
		clock_gettime(CLOCK_REALTIME, &startTime);
		ret = send_msg(g_efd, (char *)packet_buf, packet_len);
		if (ret < 0) {
			printf("send_msg failed for iteration %d, ret = %d\n", i, ret);
			goto out;
		}
		if (ret != packet_len) {
			printf("bytes written does not match send request, ret = %d, packet_len = %d\n",
				i, ret);
		    goto out;
		}

		memset(packet_buf, 0, sizeof(packet_buf));
		printf("Receiving message #%d: ", i);
		ret = recv_msg(g_efd, 256, (char *)packet_buf, &packet_len);
		if (ret < 0) {
			printf("recv_msg failed for iteration %d, ret = %d\n", i, ret);
			goto out;
		}
		clock_gettime(CLOCK_REALTIME, &endTime);
		time_ns = DIFF_NS(startTime, endTime);
		total_time += time_ns;
		avg_time = total_time / (i + 1);
		if (time_ns > max_time)
			max_time = time_ns;
		printf("test this time %llu ns, avg time %llu ns, maxtime %llu ns\n",
				time_ns, avg_time, max_time);
		usleep(500000);
		/* TODO: Verify data integrity */

		/* TODO: Reduce number of prints */
		//printf("%s\n", packet_buf);
	}

	return 0;

out:
	ret = rpmsg_char_destroy_eptdev(g_efd);
	if (ret < 0)
		perror("Can't delete the endpoint device\n");

	return ret;
}
